# 12 "src/session/strategy_parser.mll"
 
  open Session
  open Strategy

  exception SyntaxError of string

  let error f =
    Printf.kbprintf
      (fun b ->
        let s = Buffer.contents b in Buffer.clear b; raise (SyntaxError s))
      (Buffer.create 1024)
      f

  type label = {
    mutable defined: int option;
    temporary: int;
  }

  type 'a code = {
    env: 'a Session.env_session;
    mutable instr: instruction array;
    mutable next: int;
    labels: (string, label) Hashtbl.t; (* label name -> label *)
    mutable temp: int;
  }

  let create_code env =
    let h = Hashtbl.create 17 in
    Hashtbl.add h "exit" { defined = Some (-1); temporary = 0 };
    { env = env;
      instr = Array.make 10 (Igoto 0);
      next = 0;
      temp = 1;
      labels = h; }

  let enlarge_code code =
    let old = code.instr in
    let n = Array.length old in
    code.instr <- Array.make (2 * n) (Igoto 0);
    Array.blit old 0 code.instr 0 n

  let add_instr code i =
    let n = code.next in
    if n = Array.length code.instr then enlarge_code code;
    code.instr.(n) <- i;
    code.next <- n + 1

  let temp code =
    let t = code.temp in code.temp <- t + 1; t

  let define_label code l =
    let n = code.next in
    try
      let lab = Hashtbl.find code.labels l in
      if lab.defined = None then lab.defined <- Some n
      else error "duplicate label %s" l
    with Not_found ->
      let lab =  { defined = Some n; temporary = temp code } in
      Hashtbl.add code.labels l lab

  let find_label code l =
    try
      let lab = Hashtbl.find code.labels l in lab.temporary
    with Not_found ->
      let t = temp code in
      Hashtbl.add code.labels l { defined = None; temporary = t };
      t

  let prover code p =
    try
      let fp = Whyconf.parse_filter_prover p in
      Whyconf.filter_one_prover code.env.whyconf fp
    with
      | Whyconf.ProverNotFound _ ->
          error "Prover %S not installed or not configured" p
      | Whyconf.ProverAmbiguity _ ->
          error "Prover description %s is ambiguous" p

  let integer msg s =
    try int_of_string s
    with Failure _ -> error "unable to parse %s argument '%s'" msg s

  let transform code t =
    try
      ignore (Trans.lookup_transform t code.env.Session.env)
    with Trans.UnknownTrans _ ->
    try
      ignore (Trans.lookup_transform_l t code.env.Session.env)
    with Trans.UnknownTrans _->
      error "transformation %S is unknown" t


# 95 "src/session/strategy_parser.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\248\255\249\255\002\000\007\000\009\000\014\000\004\000\
    \016\000\254\255\011\000\042\000\253\255\044\000\053\000\060\000\
    \073\000\080\000\088\000\100\000\117\000\124\000\130\000\141\000\
    \167\000\194\000\179\000\151\000\244\000\003\001\008\001\016\001\
    \023\001\054\001\062\001\069\001\074\001\101\001\106\001";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\006\000\006\000\006\000\006\000\006\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\004\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\005\000";
  Lexing.lex_default = 
   "\003\000\000\000\000\000\013\000\013\000\013\000\013\000\010\000\
    \255\255\000\000\010\000\017\000\000\000\013\000\013\000\013\000\
    \013\000\017\000\022\000\013\000\013\000\013\000\022\000\255\255\
    \255\255\255\255\255\255\036\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\036\000\038\000\038\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\008\000\008\000\255\255\255\255\008\000\009\000\255\255\
    \027\000\027\000\018\000\018\000\027\000\009\000\018\000\011\000\
    \011\000\008\000\008\000\011\000\000\000\008\000\000\000\000\000\
    \008\000\000\000\255\255\007\000\000\000\255\255\000\000\027\000\
    \000\000\018\000\255\255\000\000\255\255\000\000\011\000\000\000\
    \008\000\255\255\000\000\011\000\011\000\255\255\255\255\011\000\
    \000\000\255\255\002\000\000\000\012\000\000\000\255\255\255\255\
    \000\000\012\000\255\255\012\000\000\000\255\255\255\255\000\000\
    \012\000\255\255\011\000\000\000\255\255\255\255\000\000\255\255\
    \000\000\000\000\011\000\011\000\000\000\255\255\011\000\000\000\
    \255\255\255\255\255\255\000\000\255\255\255\255\000\000\255\255\
    \000\000\018\000\018\000\005\000\255\255\018\000\012\000\006\000\
    \000\000\011\000\019\000\000\000\255\255\255\255\255\255\012\000\
    \255\255\255\255\000\000\255\255\004\000\000\000\012\000\000\000\
    \018\000\028\000\000\000\255\255\000\000\014\000\255\255\255\255\
    \000\000\000\000\255\255\012\000\255\255\018\000\018\000\255\255\
    \000\000\018\000\255\255\023\000\023\000\000\000\000\000\023\000\
    \000\000\000\000\255\255\000\000\000\000\255\255\023\000\023\000\
    \255\255\000\000\023\000\000\000\018\000\000\000\012\000\255\255\
    \027\000\027\000\023\000\000\000\027\000\255\255\000\000\000\000\
    \000\000\015\000\000\000\016\000\000\000\023\000\000\000\012\000\
    \025\000\025\000\000\000\000\000\025\000\000\000\012\000\027\000\
    \000\000\000\000\255\255\000\000\255\255\024\000\024\000\024\000\
    \024\000\024\000\024\000\024\000\024\000\024\000\024\000\025\000\
    \000\000\000\000\000\000\025\000\025\000\000\000\000\000\025\000\
    \020\000\255\255\000\000\000\000\000\000\000\000\000\000\024\000\
    \024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\
    \024\000\021\000\025\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\000\000\000\000\000\000\
    \000\000\000\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\000\000\255\255\255\255\000\000\
    \001\000\255\255\255\255\000\000\009\000\000\000\000\000\255\255\
    \000\000\255\255\000\000\009\000\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\000\000\255\255\255\255\000\000\255\255\
    \000\000\255\255\255\255\000\000\000\000\255\255\000\000\000\000\
    \255\255\255\255\000\000\255\255\255\255\000\000\255\255\000\000\
    \255\255\000\000\255\255\255\255\255\255\000\000\012\000\000\000\
    \255\255\000\000\000\000\255\255\000\000\255\255\000\000\255\255\
    \000\000\000\000\255\255\000\000\255\255\012\000\000\000\255\255\
    \255\255\000\000\012\000\255\255\000\000\000\000\000\000\255\255\
    \255\255\255\255\012\000\255\255\000\000\000\000\027\000\027\000\
    \255\255\012\000\027\000\037\000\037\000\029\000\255\255\037\000\
    \255\255\255\255\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\255\255\000\000\000\000\255\255\027\000\000\000\000\000\
    \255\255\000\000\037\000\000\000\000\000\255\255\037\000\037\000\
    \012\000\030\000\037\000\255\255\255\255\255\255\032\000\255\255\
    \012\000\000\000\000\000\031\000\255\255\000\000\000\000\012\000\
    \000\000\000\000\255\255\000\000\255\255\037\000\033\000\000\000\
    \255\255\000\000\255\255\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\255\255\000\000\000\000\000\000\
    \034\000\000\000\000\000\035\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\255\255";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\003\000\003\000\000\000\007\000\003\000\
    \004\000\004\000\005\000\005\000\004\000\010\000\005\000\006\000\
    \006\000\008\000\008\000\006\000\255\255\008\000\255\255\255\255\
    \000\000\255\255\003\000\000\000\255\255\003\000\255\255\004\000\
    \255\255\005\000\004\000\255\255\005\000\255\255\006\000\255\255\
    \008\000\006\000\255\255\011\000\011\000\013\000\013\000\011\000\
    \255\255\013\000\000\000\255\255\003\000\255\255\014\000\014\000\
    \255\255\004\000\014\000\005\000\255\255\015\000\015\000\255\255\
    \006\000\015\000\011\000\255\255\013\000\011\000\255\255\013\000\
    \255\255\255\255\016\000\016\000\255\255\014\000\016\000\255\255\
    \014\000\017\000\017\000\255\255\015\000\017\000\255\255\015\000\
    \255\255\018\000\018\000\000\000\011\000\018\000\013\000\000\000\
    \255\255\016\000\005\000\255\255\016\000\019\000\019\000\014\000\
    \017\000\019\000\255\255\017\000\000\000\255\255\015\000\255\255\
    \018\000\004\000\255\255\018\000\255\255\006\000\020\000\020\000\
    \255\255\255\255\020\000\016\000\019\000\021\000\021\000\019\000\
    \255\255\021\000\017\000\022\000\022\000\255\255\255\255\022\000\
    \255\255\255\255\018\000\255\255\255\255\020\000\023\000\023\000\
    \020\000\255\255\023\000\255\255\021\000\255\255\019\000\021\000\
    \027\000\027\000\022\000\255\255\027\000\022\000\255\255\255\255\
    \255\255\014\000\255\255\015\000\255\255\023\000\255\255\020\000\
    \024\000\024\000\255\255\255\255\024\000\255\255\021\000\027\000\
    \255\255\255\255\027\000\255\255\022\000\023\000\023\000\023\000\
    \023\000\023\000\023\000\023\000\023\000\023\000\023\000\024\000\
    \255\255\255\255\255\255\025\000\025\000\255\255\255\255\025\000\
    \019\000\027\000\255\255\255\255\255\255\255\255\255\255\024\000\
    \024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\
    \024\000\020\000\025\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\255\255\255\255\255\255\
    \255\255\255\255\025\000\025\000\025\000\025\000\025\000\025\000\
    \025\000\025\000\025\000\025\000\255\255\028\000\028\000\255\255\
    \000\000\028\000\003\000\255\255\007\000\255\255\255\255\004\000\
    \255\255\005\000\255\255\010\000\029\000\029\000\006\000\255\255\
    \029\000\030\000\030\000\255\255\028\000\030\000\255\255\028\000\
    \255\255\031\000\031\000\255\255\255\255\031\000\255\255\255\255\
    \032\000\032\000\255\255\029\000\032\000\255\255\029\000\255\255\
    \030\000\255\255\011\000\030\000\013\000\255\255\028\000\255\255\
    \031\000\255\255\255\255\031\000\255\255\014\000\255\255\032\000\
    \255\255\255\255\032\000\255\255\015\000\029\000\255\255\033\000\
    \033\000\255\255\030\000\033\000\255\255\255\255\255\255\034\000\
    \034\000\016\000\031\000\034\000\255\255\255\255\035\000\035\000\
    \017\000\032\000\035\000\036\000\036\000\028\000\033\000\036\000\
    \018\000\033\000\255\255\255\255\255\255\255\255\034\000\255\255\
    \255\255\034\000\255\255\255\255\019\000\035\000\255\255\255\255\
    \035\000\255\255\036\000\255\255\255\255\036\000\037\000\037\000\
    \033\000\029\000\037\000\038\000\038\000\020\000\031\000\038\000\
    \034\000\255\255\255\255\030\000\021\000\255\255\255\255\035\000\
    \255\255\255\255\022\000\255\255\036\000\037\000\032\000\255\255\
    \037\000\255\255\038\000\255\255\255\255\038\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\027\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\037\000\
    \255\255\255\255\255\255\255\255\038\000\255\255\255\255\255\255\
    \033\000\255\255\255\255\034\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\028\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\029\000\255\255\255\255\255\255\255\255\
    \030\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \031\000\255\255\255\255\255\255\255\255\255\255\255\255\032\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\033\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\034\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\035\000\255\255\255\255\
    \255\255\255\255\036\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\037\000\255\255\255\255\
    \255\255\255\255\038\000";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\002\000\007\000\000\000\
    \000\000\000\000\000\000\009\000\000\000\000\000\000\000\000\000\
    \014\000\000\000\016\000\000\000\000\000\027\000\034\000\045\000\
    \055\000\040\000\000\000\104\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\047\000\106\000\111\000\000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\010\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\025\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\042\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\013\000\000\000\000\000\000\000\013\000\000\000\
    \000\000\000\000\000\000\036\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\036\000\000\000\000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\001\000\004\000\004\000\001\000\000\000\004\000\
    \007\000\007\000\007\000\007\000\007\000\000\000\007\000\007\000\
    \007\000\004\000\004\000\007\000\000\000\004\000\000\000\000\000\
    \001\000\000\000\004\000\000\000\004\000\004\000\000\000\007\000\
    \004\000\007\000\000\000\016\000\016\000\000\000\007\000\016\000\
    \004\000\022\000\022\000\000\000\000\000\022\000\016\000\016\000\
    \001\000\001\000\016\000\004\000\001\000\000\000\000\000\000\000\
    \022\000\022\000\016\000\000\000\022\000\000\000\000\000\000\000\
    \022\000\000\000\000\000\000\000\000\000\016\000\000\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\022\000\
    \000\000\000\000\000\000\000\000\000\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\001\000\001\000\039\000\039\000\001\000\000\000\039\000\
    \039\000\039\000\000\000\000\000\039\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\000\000\039\000\000\000\000\000\000\000\000\000\039\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\004\000\004\000\005\000\005\000\004\000\255\255\005\000\
    \006\000\006\000\011\000\011\000\006\000\255\255\011\000\016\000\
    \016\000\018\000\018\000\016\000\255\255\018\000\255\255\255\255\
    \004\000\255\255\005\000\255\255\021\000\021\000\255\255\006\000\
    \021\000\011\000\255\255\022\000\022\000\255\255\016\000\022\000\
    \018\000\025\000\025\000\018\000\255\255\025\000\023\000\023\000\
    \035\000\035\000\023\000\021\000\035\000\255\255\255\255\255\255\
    \024\000\024\000\022\000\255\255\024\000\022\000\255\255\255\255\
    \025\000\255\255\018\000\255\255\255\255\023\000\255\255\035\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\024\000\
    \255\255\255\255\255\255\255\255\022\000\023\000\023\000\023\000\
    \023\000\023\000\023\000\023\000\023\000\023\000\023\000\024\000\
    \024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\
    \024\000\027\000\027\000\036\000\036\000\027\000\255\255\036\000\
    \037\000\037\000\255\255\255\255\037\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \027\000\255\255\036\000\027\000\255\255\036\000\255\255\037\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\027\000\255\255\036\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \018\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\022\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \027\000\255\255\036\000\255\255\255\255\255\255\255\255\255\255\
    ";
  Lexing.lex_code = 
   "\255\011\255\255\006\255\255\005\255\255\000\005\255\007\255\255\
    \008\255\255\009\255\255\010\255\255\002\008\003\009\000\006\001\
    \007\004\010\255\012\255\255\013\255\255\000\011\001\012\002\013\
    \255";
}

let rec scan code lexbuf =
  lexbuf.Lexing.lex_mem <- Array.create 14 (-1) ;   __ocaml_lex_scan_rec code lexbuf 0
and __ocaml_lex_scan_rec code lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 114 "src/session/strategy_parser.mll"
      ( scan code lexbuf )
# 401 "src/session/strategy_parser.ml"

  | 1 ->
# 116 "src/session/strategy_parser.mll"
      ( scan code lexbuf )
# 406 "src/session/strategy_parser.ml"

  | 2 ->
let
# 117 "src/session/strategy_parser.mll"
             id
# 412 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_curr_pos + -1) in
# 118 "src/session/strategy_parser.mll"
      ( define_label code id;
        scan code lexbuf )
# 417 "src/session/strategy_parser.ml"

  | 3 ->
let
# 120 "src/session/strategy_parser.mll"
                          id
# 423 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 121 "src/session/strategy_parser.mll"
      ( add_instr code (Igoto (find_label code id));
        scan code lexbuf )
# 428 "src/session/strategy_parser.ml"

  | 4 ->
let
# 123 "src/session/strategy_parser.mll"
                          p
# 434 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 123 "src/session/strategy_parser.mll"
                                                t
# 439 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(3)
and
# 123 "src/session/strategy_parser.mll"
                                                                      m
# 444 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(4) lexbuf.Lexing.lex_curr_pos in
# 124 "src/session/strategy_parser.mll"
      ( let p = prover code p in
        let t = integer "timelimit" t in
        if t <= 0 then error "timelimit %d is invalid" t;
        let m = integer "memlimit" m in
        if m <= 0 then error "memlimit %d is invalid" m;
        add_instr code (Icall_prover (p.Whyconf.prover, t, m));
        scan code lexbuf )
# 454 "src/session/strategy_parser.ml"

  | 5 ->
let
# 131 "src/session/strategy_parser.mll"
                               t
# 460 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 131 "src/session/strategy_parser.mll"
                                                   l
# 465 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_curr_pos in
# 132 "src/session/strategy_parser.mll"
      ( transform code t;
        add_instr code (Itransform (t, find_label code l));
        scan code lexbuf )
# 471 "src/session/strategy_parser.ml"

  | 6 ->
let
# 135 "src/session/strategy_parser.mll"
         c
# 477 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 136 "src/session/strategy_parser.mll"
      ( let i = Lexing.lexeme_start lexbuf in
        error "syntax error on character '%c' at position %d" c i )
# 482 "src/session/strategy_parser.ml"

  | 7 ->
# 139 "src/session/strategy_parser.mll"
      ( () )
# 487 "src/session/strategy_parser.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; __ocaml_lex_scan_rec code lexbuf __ocaml_lex_state

;;

# 141 "src/session/strategy_parser.mll"
 

  let parse env s =
    let code = create_code env in
    scan code (Lexing.from_string s);
    let label = Array.create code.temp 0 in
    let fill name lab = match lab.defined with
      | None -> error "label '%s' is undefined" name
      | Some n -> label.(lab.temporary) <- n in
    Hashtbl.iter fill code.labels;
    let solve = function
      | Icall_prover _ as i -> i
      | Itransform (t, n) -> Itransform (t, label.(n))
      | Igoto n -> Igoto label.(n) in
    Array.map solve (Array.sub code.instr 0 code.next)


# 511 "src/session/strategy_parser.ml"
